/*
* ----------------------------------------------------------------
* Copyright c                  Realtek Semiconductor Corporation, 2003  
* All rights reserved.
* 
* $Header: /home/cvsroot/uClinux-dist/linux-2.4.x/drivers/net/re865x/crypto/865xb/cryptoCmd.c,v 1.4 2004/07/12 05:16:09 yjlou Exp $
*
* Abstract: Crypto engine access command driver source code.
* $Author: yjlou $
* $Log: cryptoCmd.c,v $
* Revision 1.4  2004/07/12 05:16:09  yjlou
* *: fixed the bug of getting byte count parameter of 'init' (crypto nd hash)
*
* Revision 1.3  2004/06/23 15:36:12  yjlou
* *: removed duplicated code, and implemented original functions with the generic function:
*    rtl8651b_authEngine_md5(), rtl8651b_authEngine_sha1(),
*    rtl8651b_authEngine_hmacMd5(), rtl8651b_authEngine_hmacSha1(),
*    rtl8651b_cryptoEngine_ecb_encrypt(), rtl8651b_cryptoEngine_3des_ecb_encrypt(),
*    rtl8651b_cryptoEngine_cbc_encrypt(), rtl8651b_cryptoEngine_3des_cbc_encrypt()
* -: remove rtl8651b_cryptoEngine_cbc_encryptEmbIV() and rtl8651b_cryptoEngine_ede_cbc_encryptEmbIV()
*
* Revision 1.2  2004/06/23 10:15:45  yjlou
* *: convert DOS format to UNIX format
*
* Revision 1.1  2004/06/23 09:18:57  yjlou
* +: support 865xB CRYPTO Engine
*   +: CONFIG_RTL865XB_EXP_CRYPTOENGINE
*   +: basic encry/decry functions (DES/3DES/SHA1/MAC)
*   +: old-fashion API (should be removed in next version)
*   +: batch functions (should be removed in next version)
*
* Revision 1.10  2003/09/30 12:23:54  jzchen
* Add random test
*
* Revision 1.9  2003/09/29 09:04:51  jzchen
* Add authentication batch command
*
* Revision 1.8  2003/09/25 10:00:29  jzchen
* Add batch testing code
*
* Revision 1.7  2003/09/23 02:21:43  jzchen
* Change layout of crypto and authenticaiton engine command
*
* Revision 1.6  2003/09/18 14:33:36  jzchen
* Add throughput evaluation commands
*
* Revision 1.5  2003/09/09 02:52:14  jzchen
* Add random test to comare software HMAC and simulator HMAC
*
* Revision 1.4  2003/09/02 14:59:05  jzchen
* Execute HMAC test
*
* Revision 1.3  2003/09/01 02:51:09  jzchen
* Porting test code for MD5 and SHA-1. Test software code ready.
*
* Revision 1.2  2003/08/28 02:51:30  jzchen
* Add crypto engine comparision test
*
* Revision 1.1  2003/08/21 07:06:12  jzchen
* Using openssl des testing code to test software des function
*
*
* ---------------------------------------------------------------
*/

#include <rtl_types.h>
#include <rtl_cle.h> 
#include <cryptoCmd.h>
#include <authTest.h>
#include <rtl_glue.h>

int32 _rtl8651b_cryptoCmd(uint32 userId,  int32 argc,int8 **saved) {
	int32 size;
	int8 *nextToken;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 

	if(!strcmp(nextToken, "init")){
		uint32 descNum;
		uint32 BurstSize;
		
		cle_getNextCmdToken(&nextToken,&size,saved); 
		descNum = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		BurstSize = U32_value(nextToken);	
		if( BurstSize == 32 ) 
		{	
			rtlglue_printf("Initialize crypto engine driver with %u descriptors working on 32-byte mode\n", descNum);
			rtl8651b_cryptoEngine_init(descNum, TRUE);
		}
		else 
		{	
			rtlglue_printf("Initialize crypto engine driver with %u descriptors working on 16-byte mode\n", descNum);
			rtl8651b_cryptoEngine_init(descNum, FALSE);
		}
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "sim")){
		rtlglue_printf("Run des and des simulator test\n");
		destest();
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "rand")){
		uint32 randSeed, roundLimit;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		randSeed = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		roundLimit = U32_value(nextToken);	
		rtlglue_printf("Run DES random test seed %u and %u rounds\n", randSeed, roundLimit);
		runDes8651bGeneralApiRandTest(randSeed, roundLimit);
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "asic")){
		uint32 round, funStart, funEnd, lenStart, lenEnd, offsetStart, offsetEnd;
		rtlglue_printf("Run des simulator and 8651b crypto engine test\n");
		cle_getNextCmdToken(&nextToken,&size,saved); 
		round = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		funStart = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		funEnd = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		lenStart = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		lenEnd = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		offsetStart = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		offsetEnd = U32_value(nextToken);	
		rtlglue_printf("Round %u Function %u - %u Len %u - %u Offset %u - %u\n", round, funStart, funEnd, lenStart, lenEnd, offsetStart, offsetEnd);
//		des8651bTest(round, funStart, funEnd, lenStart, lenEnd, offsetStart, offsetEnd);
		runDes8651bGeneralApiTest(round, funStart, funEnd, lenStart, lenEnd, offsetStart, offsetEnd);
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "throughput")) {
		uint32 round, startMode, endMode, pktLen;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		 if(!strcmp(nextToken, "asic")) {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			round = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			startMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			endMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			pktLen = U32_value(nextToken);	
		 	des8651bAsicThroughput(round, startMode, endMode, pktLen);
			return SUCCESS;
		 }
		 else if(!strcmp(nextToken, "sw")) {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			round = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			startMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			endMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			pktLen = U32_value(nextToken);	
		 	des8651bSwThroughput(round, startMode, endMode, pktLen);
			return SUCCESS;
		 }
	}

	return FAILED;
}

int32 _rtl8651b_authenticationCmd(uint32 userId,  int32 argc,int8 **saved) {
	int32 size;
	int8 *nextToken;
	
	cle_getNextCmdToken(&nextToken,&size,saved); 

	if(!strcmp(nextToken, "init")){
		uint32 descNum;
		uint32 BurstSize;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		descNum = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		BurstSize = U32_value(nextToken);	
		if(BurstSize == 32) 
		{	
			rtlglue_printf("Initialize authentication engine driver with %u descriptors working on 32-byte mode\n", descNum);
			rtl8651b_authEngine_init(descNum, TRUE);
		}
		else 
		{	
			rtlglue_printf("Initialize authentication engine driver with %u descriptors working on 16-byte mode\n", descNum);
			rtl8651b_authEngine_init(descNum, FALSE);
		}
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "sim")){
		rtlglue_printf("Run MD5 and MD5 simulator test\n");
		md5test();
		rtlglue_printf("Run SHA-1 and SHA-1 simulator test\n");
		sha1test();
		rtlglue_printf("Run HMAC MD5 and HMAC MD5 simulator test\n");
		hmactest();
		rtlglue_printf("Run HMAC random test\n");
		hmacSimRandTest(2);
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "rand")){
		uint32 randSeed, roundLimit;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		randSeed = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		roundLimit = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		rtlglue_printf("Run %s random test seed %u for %u rounds\n",nextToken, randSeed, roundLimit);
		if(!strcmp(nextToken, "asic"))
			runAuth8651bGeneralApiRandTest(randSeed, roundLimit);
		else
			hmacSimRandTest(roundLimit);
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "asic")) {
		uint32 round, funStart, funEnd, lenStart, lenEnd, keyLenStart, keyLenEnd, offsetStart, offsetEnd;
		rtlglue_printf("Run auth simulator and 8651b auth engine test\n");
		cle_getNextCmdToken(&nextToken,&size,saved); 
		round = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		funStart = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		funEnd = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		lenStart = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		lenEnd = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		keyLenStart = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		keyLenEnd = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		offsetStart = U32_value(nextToken);	
		cle_getNextCmdToken(&nextToken,&size,saved); 
		offsetEnd = U32_value(nextToken);	
		rtlglue_printf("Round %u Function %u - %u Len %u - %u keyLen %u - %u Offset %u - %u \n", round, funStart, funEnd, lenStart, lenEnd, keyLenStart, keyLenEnd, offsetStart, offsetEnd);
		runAuth8651bGeneralApiTest(round, funStart, funEnd, lenStart, lenEnd, keyLenStart, keyLenEnd, offsetStart, offsetEnd);
		return SUCCESS;
	}
	else if(!strcmp(nextToken, "throughput")) {
		uint32 round, startMode, endMode, pktLen;
		cle_getNextCmdToken(&nextToken,&size,saved); 
		 if(!strcmp(nextToken, "asic")) {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			round = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			startMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			endMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			pktLen = U32_value(nextToken);	
		 	auth8651bAsicThroughput(round, startMode, endMode, pktLen);
			return SUCCESS;
		 }
		 else if(!strcmp(nextToken, "sw")) {
			cle_getNextCmdToken(&nextToken,&size,saved); 
			round = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			startMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			endMode = U32_value(nextToken);	
			cle_getNextCmdToken(&nextToken,&size,saved); 
			pktLen = U32_value(nextToken);	
		 	auth8651bSwThroughput(round, startMode, endMode, pktLen);
			return SUCCESS;
		 }
	}
	return FAILED;
}


cle_exec_t rtl865x_crypt_cmds[] = 
{
	{   "crypto",
		"Crypto engine commands",
		" { init %d'descNum' | sim |"
		"   rand %d'seed' %d'round' |"
		"   asic %d'round' %d'funStart' %d'funEnd' %d'lenStart' %d'lenEnd' %d'offsetStart' %d'offsetEnd' |"
		"   batch %d'round' %d'funStart' %d'funEnd' %d'lenStart' %d'lenEnd' %d'offsetStart' %d'offsetEnd' %d'batchStart' %d'batchEnd' |"
		"   throughput { asic | sw } %d'round' %d'StartMode' %d'EndMode' %d'PktLen' }",
		_rtl8651b_cryptoCmd,
		CLE_USECISCOCMDPARSER,
		NULL,
		NULL
	},
	{   "auth",
		"Auth engine commands",
		" { init %d'descNum' | sim |"
		"   rand %d'seed' %d'round' { asic | sw } |"
		"   asic %d'round' %d'funStart' %d'funEnd' %d'lenStart' %d'lenEnd' %d'keyStart' %d'keyEnd' %d'offsetStart' %d'offsetEnd' |"
		"   batch %d'round' %d'funStart' %d'funEnd' %d'lenStart' %d'lenEnd' %d'keyStart' %d'keyEnd' %d'offsetStart' %d'offsetEnd' %d'batchStart' %d'batchEnd' |"
		"   throughput { asic | sw } %d'round' %d'StartMode' %d'EndMode' %d'PktLen' }",
		_rtl8651b_authenticationCmd,
		CLE_USECISCOCMDPARSER,
		NULL,
		NULL
	},
};
